/*Elkulator v0.7 by Tom Walker
  Cycle-accurate ULA emulation*/
#include "sdlalleg.h"
#include "elk.h"

unsigned short pc;
int bitcount=0,sbitcount=9;
int tapeout=0;
int tapeoutwait=0;
#define INT_VBL      0x04
#define INT_RTC      0x08
#define INT_RECEIVE  0x10
#define INT_SEND     0x20
#define INT_HIGHTONE 0x40

#define BIT_LENGTH 1628

unsigned short tapedat;
int tapewrite=0;
int irq=0,nmi=0;
int extrom,rombank,intrombank;
int tapeon;

int soundlimit,soundon,soundcount,soundstat;

struct
{
        unsigned char isr,ier;
        unsigned char addrlo,addrhi;
        unsigned char recdat;
        
        int x,y;
        int dispon;
        int oddeven;
        int mode;
        
        unsigned short addr,addrback;
        int sc;
        int framecount;
        int draw;
        
        unsigned char tapelatch;
} ula;

BITMAP *b,*b16;

unsigned char pal[16];
int palwritenum=0,palwritenum2=0;
PALETTE elkpal =
{
      {0,0,0},
      {63,0,0},
      {0,63,0},
      {63,63,0},
      {0,0,63},
      {63,0,63},
      {0,63,63},
      {63,63,63},
};

unsigned char ulalookup[256];
void initula()
{
        int c;
//        allegro_init();
        set_gfx_mode(GFX_AUTODETECT_WINDOWED,640,512,0,0);
        b=create_bitmap(640,313);
        b16=create_bitmap(640*2,512);
        set_palette(elkpal);
        for (c=0;c<256;c++)
        {
                ulalookup[c]=0;
                if (c&2) ulalookup[c]|=1;
                if (c&8) ulalookup[c]|=2;
                if (c&0x20) ulalookup[c]|=4;
                if (c&0x80) ulalookup[c]|=8;
        }
}

void resetula()
{
        ula.x=ula.y=ula.sc=0;
        ula.dispon=ula.oddeven=0;
        ula.isr=2;
        ula.ier=0;
        ula.mode=6;
        tapeon=0;
        ula.framecount++;
        soundon=soundstat=0x7F;
        tapewrite=0;
        wipesoundbuffer();
}

void updateulaints()
{
        if (ula.isr&ula.ier&0x7C)
        {
                ula.isr|=1;
                irq=1;
//                printf("Interrupt %02X %02X\n",ula.isr,ula.ier);
        }
        else
        {
                ula.isr&=~1;
                irq=0;
        }
}

int hightone=0;

void intula(unsigned char num)
{
        if (num==INT_HIGHTONE)
        {
                hightone=1;
//                printf("In high tone!\n");
                return;
        }
//        if (num&0x10) printf("INT &0x10\n");
        ula.isr|=num;
        updateulaints();
}

void writeula(unsigned short addr, unsigned char val)
{
        switch (addr&0xF)
        {
                case 0: /*Interrupt control*/
                ula.ier=val;
                updateulaints();
                break;
                case 2:
                ula.addrlo=val&0xE0;
                break;
                case 3:
                ula.addrhi=val&0x3F;
                break;
                case 4:
//                printf("Write tape out %i %i %02X\n",tapeoutwait,tapewrite,val);
                if (tapewrite<BIT_LENGTH)
                {
                        tapewrite=BIT_LENGTH;
                        ula.isr&=~INT_SEND;
                        updateulaints();
                }
//                ula.isr&=~INT_RECEIVE;
//                updateulaints();
                sbitcount=9;
                tapeoutwait=0;
                bitcount=0;
                tapedat=(val<<2)|1;
//                tapewrite=2000000/(1200/10);
                break;
                case 5: /*ROM control / interrupt clear*/
                if (val&0x10) ula.isr&=~INT_VBL;
                if (val&0x20) ula.isr&=~INT_RTC;
                if (val&0x40) ula.isr&=~INT_HIGHTONE;
                updateulaints();
                rombank=val&0xF;
                if (rombank>=0xC) extrom=1;
                if ((rombank&0xC)==8)
                {
                        extrom=0;
                        intrombank=rombank;
                }
                break;
                case 6: /*Timer*/
                soundlimit=(val<<16);
                break;
                case 7: /*Misc control*/
                if ((((val&6)==2)?1:0)!=soundon)
                   soundon=((val&6)==2)?1:0;
                ula.mode=(val>>3)&7;
                if (tapeon && !(val&64)) resetsound();
                tapeon=val&64;
                tapeout=((val&6)==4);
                break;
                case 0x8: case 0xA: case 0xC: case 0xE:
                switch ((addr>>1)&3)
                {
                        case 0:
                        pal[0]&=~4;
                        pal[2]&=~4;
                        pal[8]&=~6;
                        pal[10]&=~6;
                        if (!(val&0x10)) pal[0]|=4;
                        if (!(val&0x20)) pal[2]|=4;
                        if (!(val&0x40)) pal[8]|=4;
                        if (!(val&0x80)) pal[10]|=4;
                        if (!(val&0x4)) pal[8]|=2;
                        if (!(val&0x8)) pal[10]|=2;
                        break;
                        case 1:
                        pal[4]&=~4;
                        pal[6]&=~4;
                        pal[12]&=~6;
                        pal[14]&=~6;
                        if (!(val&0x10)) pal[4]|=4;
                        if (!(val&0x20)) pal[6]|=4;
                        if (!(val&0x40)) pal[12]|=4;
                        if (!(val&0x80)) pal[14]|=4;
                        if (!(val&0x4)) pal[12]|=2;
                        if (!(val&0x8)) pal[14]|=2;
                        break;
                        case 2:
                        pal[5]&=~4;
                        pal[7]&=~4;
                        pal[13]&=~6;
                        pal[15]&=~6;
                        if (!(val&0x10)) pal[5]|=4;
                        if (!(val&0x20)) pal[7]|=4;
                        if (!(val&0x40)) pal[13]|=4;
                        if (!(val&0x80)) pal[15]|=4;
                        if (!(val&0x4)) pal[13]|=2;
                        if (!(val&0x8)) pal[15]|=2;
                        break;
                        case 3:
                        pal[1]&=~4;
                        pal[3]&=~4;
                        pal[9]&=~6;
                        pal[11]&=~6;
                        if (!(val&0x10)) pal[1]|=4;
                        if (!(val&0x20)) pal[3]|=4;
                        if (!(val&0x40)) pal[9]|=4;
                        if (!(val&0x80)) pal[11]|=4;
                        if (!(val&0x4)) pal[9]|=2;
                        if (!(val&0x8)) pal[11]|=2;
                        break;
                }
                break;
                case 0x9: case 0xB: case 0xD: case 0xF:
                switch ((addr>>1)&3)
                {
                        case 0:
                        pal[0]&=~3;
                        pal[2]&=~3;
                        pal[8]&=~1;
                        pal[10]&=~1;
                        if (!(val&1)) pal[0]|=1;
                        if (!(val&2)) pal[2]|=1;
                        if (!(val&4)) pal[8]|=1;
                        if (!(val&8)) pal[10]|=1;
                        if (!(val&16)) pal[0]|=2;
                        if (!(val&32)) pal[2]|=2;
                        break;
                        case 1:
                        pal[4]&=~3;
                        pal[6]&=~3;
                        pal[12]&=~1;
                        pal[14]&=~1;
                        if (!(val&1)) pal[4]|=1;
                        if (!(val&2)) pal[6]|=1;
                        if (!(val&4)) pal[12]|=1;
                        if (!(val&8)) pal[14]|=1;
                        if (!(val&16)) pal[4]|=2;
                        if (!(val&32)) pal[6]|=2;
                        break;
                        case 2:
                        pal[5]&=~3;
                        pal[7]&=~3;
                        pal[13]&=~1;
                        pal[15]&=~1;
                        if (!(val&1)) pal[5]|=1;
                        if (!(val&2)) pal[7]|=1;
                        if (!(val&4)) pal[13]|=1;
                        if (!(val&8)) pal[15]|=1;
                        if (!(val&16)) pal[5]|=2;
                        if (!(val&32)) pal[7]|=2;
                        break;
                        case 3:
                        pal[1]&=~3;
                        pal[3]&=~3;
                        pal[9]&=~1;
                        pal[11]&=~1;
                        if (!(val&1)) pal[1]|=1;
                        if (!(val&2)) pal[3]|=1;
                        if (!(val&4)) pal[9]|=1;
                        if (!(val&8)) pal[11]|=1;
                        if (!(val&16)) pal[1]|=2;
                        if (!(val&32)) pal[3]|=2;
                        break;
                }
                break;
        }
}

unsigned char readula(unsigned short addr)
{
        unsigned char temp;
        switch (addr&0xF)
        {
                case 0: /*Interrupt status*/
                temp=ula.isr|0x80;
                ula.isr&=~2;
/*                if (temp&0x10)
                {
                        printf("Reading int %04X\n",pc);
                }*/
                return temp;
                case 4:
                ula.isr&=~0x10;
                updateulaints();
//                printf("Read tapelatch %02X\n",ula.tapelatch);
                return ula.tapelatch;
//                return ula.recdat;
        }
}

unsigned char taperecdat;

void receive(unsigned char val)
{
        ula.recdat=taperecdat=val;
        hightone=0;
        bitcount=9;
//        printf("Receive dat %02X\n",ula.recdat);
/*        if (reallyfasttape && inreallyfasttape)
        {
//                rpclog("Recieve really fast data %02X\n",val);
                reallyfasttapebreak=1;
//                ram[0xC0]=0x80;
//                ram[0xBD]=val;
        }*/
//        else
//        {
//        intula(INT_RECEIVE);
//        }
}

void tapenextbyte()
{
        intula(INT_SEND);
}

int fasttapebreak;
int pauseit=0;
int cswena;
int bitcount;
void polltape()
{
        fasttapebreak=0;
        tapewrite+=BIT_LENGTH;
        if (pauseit)
        {
                if (!tapeout)
                {
                        if (tapeon)
                        {
                                if (cswena) pollcsw();
                                else        polluef();
                        }
                }
                return;
        }
        sbitcount--;
        if (tapeoutwait==1) tapeoutwait=2;
        if (sbitcount<=0)
        {
                sbitcount=9;
                ula.isr|=INT_SEND;
                updateulaints();
                if (!tapeoutwait) tapeoutwait=1;
        }

        if (bitcount>0)
        {
                bitcount--;
                if (bitcount==7)
                {
                        ula.isr&=~INT_RECEIVE;
                        updateulaints();
                }
        }
        else
        {
                if ((tapedat&3)==1)
                {
                        ula.isr&=~INT_HIGHTONE;
                        ula.isr|=INT_RECEIVE;
                        updateulaints();
                        ula.tapelatch=ula.recdat;
                        bitcount=9;
                        fasttapebreak=1;
//                        printf("Receive int %02X\n",ula.tapelatch);
                }
                if ((tapedat&0x3FF)==0x3FF)
                {
                        ula.isr|=INT_HIGHTONE;
                        updateulaints();
                        bitcount=0;
                        fasttapebreak=1;
                }
                if (!tapeout)
                {
                        if (tapeon)
                        {
                                if (cswena) pollcsw();
                                else        polluef();
                        }
                }
        }

        if (hightone || tapeout) tapedat=(tapedat>>1)|(1<<9);
        else if (!tapeon) tapedat>>=1;
        else
        {
//                printf("%i - was %03X ",bitcount,tapedat&0x3FF);
                if (bitcount==9)      tapedat=(tapedat>>1)|(1<<9);
                else if (bitcount==8) tapedat>>=1;
                else
                {
                        tapedat=(tapedat>>1)|((taperecdat&1)<<9);
                        taperecdat>>=1;
                }
//                printf("now %03X\n",tapedat&0x3FF);
        }
}

void reallyfasttapepoll()
{
        int c;
        for (c=0;c<1000;c++)
        {
                polltape();
                if (fasttapebreak) break;
        }
}
int modescs[8]={8,8,8,10,8,8,10,10};
unsigned short modelens[8]={0x5000,0x5000,0x5000,0x4000,0x2800,0x2800,0x2000,0x2000};
int modeend[8]={256,256,256,250,256,256,250,250};

int nextulapoll;
int numlines=0;
void yield()
{
        unsigned char temp;
        int x,c;
        unsigned short tempaddr;
        int col;
        int oldcycles;
//        if (nextulapoll) printf("Beginning poll %i ",ula.x);
        while (ulacycles<cycles)
        {
                oldcycles=ulacycles;
                while (ula.x<640 && ula.dispon && ula.draw && (ulacycles<cycles))
                {
                        if (ula.sc&8)
                        {
                                for (x=0;x<8;x++)
                                    b->line[ula.y][ula.x+x]=0;
                        }
                        else if (!(ula.x&8) || !(ula.mode&4))
                        {
                                if (ula.mode&4) tempaddr=ula.addr+ula.sc+((ula.x>>1)&~7);
                                else            tempaddr=ula.addr+ula.sc+(ula.x&~7);
                                if (tempaddr&0x8000) tempaddr-=modelens[ula.mode];
                                temp=ram[tempaddr];
                                if (HALFSIZE)
                                {
                                        switch (ula.mode)
                                        {
                                                case 0: case 3:
                                                for (x=0;x<8;x++)
                                                {
                                                        col=ulalookup[temp&0x80];
                                                        b->line[ula.y][(ula.x+x)>>1]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                                case 1:
                                                for (x=0;x<8;x+=2)
                                                {
                                                        col=ulalookup[temp&0x88];
                                                        b->line[ula.y][(ula.x+x)>>1]=pal[col];
//                                                        b->line[ula.y][ula.x+x+1]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                                case 2:
                                                for (x=0;x<8;x+=4)
                                                {
                                                        col=ulalookup[temp];
                                                        b->line[ula.y][(ula.x+x)>>1]=pal[col];
//                                                        b->line[ula.y][ula.x+x+1]=pal[col];
                                                        b->line[ula.y][(ula.x+x+2)>>1]=pal[col];
//                                                        b->line[ula.y][ula.x+x+3]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                                case 4: case 6: case 7:
                                                for (x=0;x<16;x+=2)
                                                {
                                                        col=ulalookup[temp&0x80];
                                                        b->line[ula.y][(ula.x+x)>>1]=pal[col];
//                                                        b->line[ula.y][ula.x+x+1]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                                case 5:
                                                for (x=0;x<16;x+=4)
                                                {
                                                        col=ulalookup[temp&0x88];
                                                        b->line[ula.y][(ula.x+x)>>1]=pal[col];
//                                                        b->line[ula.y][ula.x+x+1]=pal[col];
                                                        b->line[ula.y][(ula.x+x+2)>>1]=pal[col];
//                                                        b->line[ula.y][ula.x+x+3]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                        }
                                }
                                else
                                {
                                        switch (ula.mode)
                                        {
                                                case 0: case 3:
                                                for (x=0;x<8;x++)
                                                {
                                                        col=ulalookup[temp&0x80];
                                                        b->line[ula.y][ula.x+x]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                                case 1:
                                                for (x=0;x<8;x+=2)
                                                {
                                                        col=ulalookup[temp&0x88];
                                                        b->line[ula.y][ula.x+x]=pal[col];
                                                        b->line[ula.y][ula.x+x+1]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                                case 2:
                                                for (x=0;x<8;x+=4)
                                                {
                                                        col=ulalookup[temp];
                                                        b->line[ula.y][ula.x+x]=pal[col];
                                                        b->line[ula.y][ula.x+x+1]=pal[col];
                                                        b->line[ula.y][ula.x+x+2]=pal[col];
                                                        b->line[ula.y][ula.x+x+3]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                                case 4: case 6: case 7:
                                                for (x=0;x<16;x+=2)
                                                {
                                                        col=ulalookup[temp&0x80];
                                                        b->line[ula.y][ula.x+x]=pal[col];
                                                        b->line[ula.y][ula.x+x+1]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                                case 5:
                                                for (x=0;x<16;x+=4)
                                                {
                                                        col=ulalookup[temp&0x88];
                                                        b->line[ula.y][ula.x+x]=pal[col];
                                                        b->line[ula.y][ula.x+x+1]=pal[col];
                                                        b->line[ula.y][ula.x+x+2]=pal[col];
                                                        b->line[ula.y][ula.x+x+3]=pal[col];
                                                        temp<<=1;
                                                }
                                                break;
                                        }
                                }
                        }
                        ula.x+=8;
                        ulacycles++;
                }
                while (ula.x<640 && ula.y<256 && !ula.dispon && ula.draw && (ulacycles<cycles))
                {
                        for (x=0;x<8;x++)
                            b->line[ula.y][ula.x+x]=0;
                        ula.x+=8;
                        ulacycles++;
                }
/*                if (ula.x==640 && ula.y==99)
                {
                        ula.isr|=INT_RTC;
                        updateulaints();
                }*/
                if (ulacycles==oldcycles)
                {
                        ula.x+=8;
                        numlines+=8;
                        if (ula.x==1024)
                        {
                                /*Sound handling*/
                                if (!tapeon || !tapespeed)
                                {
                                        for (c=0;c<2;c++)
                                        {
                                                soundcount+=0x20000;
                                                if (soundcount>=soundlimit)
                                                {
                                                        soundcount-=soundlimit;
                                                        soundstat^=0x7F;
                                                }
                                                if (soundlimit<0x20000) addsound((soundon)?0x7F:0);
                                                else                    addsound((soundon)?soundstat:0);
                                        }
                                }
                                
                                ula.x=0;
                                ula.y++;
                                if (ula.dispon)
                                {
                                        ula.sc++;
                                        if (ula.sc==modescs[ula.mode])
                                        {
                                                ula.sc=0;
                                                ula.addr+=((ula.mode&4)?320:640);
                                                /*This does seem odd, that the ULA would fix up
                                                  the address at this point. But Firetrack has
                                                  graphic problems if this is not done*/
                                                if (ula.addr&0x8000) ula.addr-=modelens[ula.mode];
                                                ula.addrback=ula.addr;
                                        }
                                        ula.addr=ula.addrback;
                                }
                                if (ula.y==99)
                                {
                                        ula.isr|=INT_RTC;
                                        updateulaints();
                                }
                                c=modeend[ula.mode];
                                if (ula.oddeven) c++;
                                if (ula.y==c)
                                {
                                        ula.dispon=0;
                                        ula.isr|=INT_VBL;
                                        updateulaints();
                                }
                                if (ula.y==258) ula.dispon=0;
                                if (ula.y==((ula.oddeven)?313:312) || ula.y>=313)
                                {
//                                        printf("Numlines %i\n",numlines/1024);
                                        numlines=0;
                                        ula.y=0;
                                        ula.oddeven^=1;
                                        ula.dispon=1;
                                        ula.sc=0;
                                        ula.addr=(ula.addrlo<<1)|(ula.addrhi<<9);
//                                        printf("%04X\n",ula.addr);
                                        if (!ula.addr) ula.addr=0x8000-modelens[ula.mode];
                                        
                                        if (ula.draw)
                                        {
                                                switch (drawmode)
                                                {
                                                        case SCANLINES:
                                                        blitscan(b,screen,0,0,0,0,640,512);
                                                        break;
                                                        case LINEDBL:
                                                        blitdbl(b,screen,0,0,0,0,640,512);
                                                        break;
                                                        case _2XSAI:
                                                        blit2xsai(b,b16,320,256);
                                                        blit(b16,screen,0,0,0,0,640,512);
                                                        break;
                                                        case SCALE2X:
                                                        blitscale2x(b,320,256);
                                                        break;
                                                        case EAGLE:
                                                        bliteagle(b,b16,320,256);
                                                        blit(b16,screen,0,0,0,0,640,512);
                                                        break;
                                                        case PAL:
//                                                        blitfilter(b,b16,0,0,0,0,640,256);
                                                        palfilter(b,b16);
                                                        blitscan16(b16,screen,0,0,0,0,640,512);
                                                        break;
                                                }
                                        }
                                        wait50();
                                        ula.addrback=ula.addr;
                                        ula.framecount++;
                                        if (ula.framecount==25) ula.framecount=0;
                                        ula.draw=((!tapeon || !tapespeed) || !ula.framecount);
                                }
                        }
                        ulacycles++;
                }
        }
/*        if (nextulapoll)
        {
                nextulapoll=0;
                printf("%i\n",ula.x);
        }*/
}

void waitforramsync()
{
        int temp;
//        return;
        if (!(ula.mode&4) && ula.dispon && ula.x<640)
        {
                temp=640-ula.x;
                cycles+=(temp/8);
//                nextulapoll=1;
        }
        if (cycles&1) cycles++;
}
